#include "rom.hh"
#include "defs.hh"

void NopRegion(unsigned address, unsigned bytes)
{
    unsigned remain = bytes;
    unsigned char replacement[bytes];

    unsigned target = 0;

    /* If the range is small, we create JMPs. */
    while(remain > 3)
    {
        unsigned endpos = address+target+remain;
        replacement[target++] = 0x4C; // JMP
        replacement[target++] = (endpos   ) &0xFF;
        replacement[target++] = (endpos>>8) &0xFF;
        remain -= 3;
        
        unsigned freepos = address+target;
        
        fprintf(stderr, "Adding free(%06X, %u)\n", address+target, endpos-freepos);
        
        ROM.AddFreeSpace(address+target, endpos-freepos);
        remain = address+bytes - endpos;
    }

    while(target < bytes) { replacement[target++] = 0xEA; }
    ROM.WriteBytes(address, bytes, replacement);
}
void AlterDelay(unsigned fivebyteaddr, unsigned char nframes)
{
    ROM.Write(fivebyteaddr+1, nframes);
}


static struct ApuPankkiFunc
{
    static const unsigned char newbank = 0x0E;
    unsigned Addr;
    bool Ready;
    
public:
    ApuPankkiFunc() : Addr(0), Ready(false)
    {
    }

    unsigned Init(unsigned endaddr)
    {
        static const unsigned char Replace[] =
        {
            0x08, // php
            0x48, // pha
            0xA9,newbank,      // lda #newbank,
            0x85,0x42,         // sta CurrentBank
            0x8D,newbank,0xC0, // sta 0xC000 + newbank,
            0x68, // pla
            0x28, // plp
            0x60
        };
        const unsigned size = sizeof(Replace);
        
        endaddr -= size;
        ROM.WriteBytes(endaddr, size, Replace);
        
        Addr  = endaddr;
        Ready = true;
        
        return size;
    }
    
    unsigned GetAddress(unsigned nbytes) const
    {
        return ROM.FindFreeSpaceIn(newbank, nbytes);
    }
    
} ApuPankkiFunc;

void MoveCodeAway(unsigned beginaddr, unsigned endaddr,
                  bool Bank5=true)
{
    unsigned nbytes_orig = endaddr-beginaddr;
    unsigned char Buf[nbytes_orig];
    
    ROM.ReadData(beginaddr, nbytes_orig, Buf); // should include a "RTS"
    
    const unsigned newaddr = ApuPankkiFunc.GetAddress(nbytes_orig);
    
    ROM.WriteBytes( newaddr, nbytes_orig, Buf);
    fprintf(stderr, "Backupped %u bytes to %X\n", nbytes_orig, newaddr);
    
    if(!ApuPankkiFunc.Ready)
    {
        endaddr -= ApuPankkiFunc.Init(endaddr);
    }
    
    unsigned finaladdr = Bank5 ? 0xC3B3 /*SwitchBank05*/
                               : 0xC39C /*SwitchBankStage*/
                               ;
    
    const unsigned bankoffs = ApuPankkiFunc.Addr;
    const unsigned char Replace[] =
    {
        0x20, bankoffs&0xFF, bankoffs>>8, // jsr
        0x20, newaddr&0xFF, (newaddr>>8)&0xFF, // jsr
        0x4C, finaladdr&0xFF, (finaladdr>>8)&0xFF, // jmp
    };
    
    const unsigned size = sizeof(Replace);
    ROM.WriteBytes(beginaddr, size, Replace); beginaddr += size;
    
    if(beginaddr > endaddr)
    {
        std::fprintf(stderr, "MoveCodeAway: Too small region given (%X,%X)\n",
            beginaddr,endaddr);
    }
    
    fprintf(stderr, "Adding free(%06X, %u)\n", beginaddr, endaddr-beginaddr);
    ROM.AddFreeSpace(beginaddr, endaddr-beginaddr);
    while(beginaddr < endaddr) ROM.Write(beginaddr++, 0xEA);
}

void MakeSpaceInLastSeg()
{
    MoveCodeAway(0xF516, 0xF532+1);
    //MoveCodeAway(0xF7E2, 0xF825+1);
    //MoveCodeAway(0xF841, 0xF889+1);
    //MoveCodeAway(0xE003, 0xE03A+1); // BombmanStrategy0
    //MoveCodeAway(0xE076, 0xE0A1+1); // BombmanStrategy2
    
    ROM.Write(0xCD9F, 0xD0); // jmp -> BNE
    ROM.Write(0xCDA0, 68);   // +68
    ROM.Write(0xCDA1, 0xEA); // nop
    //MoveCodeAway(0xCD8A, 0xCDE7+1, false); // GetMegamanPositionInStage
}

void ExpandBlockMaps()
{
    //MakeSpaceInLastSeg();
    
    // Remap Wily stages. The bank was (stage-6), it's now (stage+1).
    ROM.Write(0xC3A7, 0xFF); // sbc -1 is equal to adc +1, and we had carry set so sbc is good.
    
    // Overwrite to banks 7,8,9,A,B
    // F = last page
    // E = for custom code, should it be needed.
    for(unsigned bank=0; bank<5; ++bank)
    {
        unsigned orig_addr = MakeNESaddr(bank,   0x0000);
        unsigned new_addr  = MakeNESaddr(7+bank, 0x0000);
        
        unsigned char Buf[0x1000];
        ROM.ReadData(orig_addr, 0x1000, Buf);
        ROM.WriteBytes(new_addr,0x1000, Buf);
        
        ROM.DelFreeSpace(new_addr,0x1000);
    }
}

void RemoveDelays()
{
    MakeSpaceInLastSeg();
    
    /* Mark the end of BANK 5 as free. */
    ROM.AddFreeSpace(0x05BF76, 0xC000-0xBF76);

    NopRegion(0xC102, 5*2); // after stage cleared, before end

    NopRegion(0xC08C, 5*4); // game or stage cleared
    NopRegion(0xC0D2, 5*1); // related to stage clearing
    NopRegion(0xC0F2, 3);   // -"-
    
    AlterDelay(0xC223, 0x01); // megaman killed
    //NopRegion(0xC223, 5*1); // megaman killed
    
    NopRegion(0xCA2D, 5*1); // boss killed

    NopRegion(0xC241, 9);    // After Megaman killed
    AlterDelay(0xD0F3, 0x02); // shutter speed
    NopRegion(0xEF84, 5*1); // before boss appears
    NopRegion(0xEFD5, 5*1); // boss pose
//    NopRegion(0xF00A, 5*1); // ? 0x80 
//    NopRegion(0xF012, 5*1); // ? 0x80 

    //AlterDelay(0xCB48, 0x10); // ? 0x80 - related to winning Wily4 battle.
    //AlterDelay(0xF203, 0x7F); // changing this number makes Wily4 battle impossible to win
    NopRegion(0xF24F, 3); // wily4 appears faster now (no palette fadings)
    
    NopRegion(0xF059, 5*1); // boss life meter filling? maybe
    AlterDelay(0x05913B, 1); // stage starts in # frames
    NopRegion(0x059290, 3); // frame wait in scrolling
    NopRegion(0x059301, 3); // frame wait in scrolling
    
    // To make scrolling faster, disable graphics updates:
    //NopRegion(0x05A044, 9); // in horizontal scrolling
    //NopRegion(0x05A082, 12); // in vertical scrolling
    
    // boss_sel:delay after shuffling numbers
    // - it's enough for it to be noticed.
    AlterDelay(0x06B945, 0x10);
    NopRegion(0x06B90C, 3); // remove the delay in displaying numbers
    
    // climbing speed
    ROM.Write(0x059829+1, 0xFE);
    ROM.Write(0x0597F7+1, 0x01);
    ROM.Write(0x059323+1, 0x01);
    
    // attempt to remove the delay in displaying boss name
    ROM.Write(0x06BB57+1, 1); // speed of boss names. Was 5, now 1.
    ROM.Write(0x06B839, 0x30); // pose length - enough for all but Cutman :)
    
    ROM.Write(0xCA4C+1, 4); // give the 'stage clear' music when boss dies (RUSH!)
    NopRegion(0xCA47,5); // ignore "stop music" when boss dies
    AlterDelay(0xCAF4, 0x10); // when Rockmonster dies (wily1)
    
    NopRegion(0xC071,5); // don't give 'stage clear' music when the item is collected
    
    // There is now no delay after killing a boss, so
    // the music that comes after it is best left supressed!
    ROM.Write(0x06B6DA+1, 0x1A); // substitute stage selection music!
    NopRegion(0x06B77A, 5); // remove boss selection music!
    //AlterDelay(0x06B775, 0x12); // change boss selection sfx
    
    NopRegion(0x06B7F6, 3); // remove the delay when boss jumps on screen
    
    AlterDelay(0x05931B, 0x08); // flashing when battle starts
    
#if 1 /* THIS SECTION ONLY WORKS IN USA MEGAMAN */
    static const unsigned char MusicPatch[] =
    {
        0xA9, 0x22, // lda sfx
        0x20, 0x77, 0xC4,  // jsr IssueSound
        0x60 // Rts
    };
    unsigned LevelEndMusicPatch = ROM.FindFreeSpaceIn(ROMmap_npages-1, sizeof(MusicPatch));
    ROM.WriteBytes(LevelEndMusicPatch, sizeof(MusicPatch), MusicPatch);
    
    ROM.Write(0xFA19, 0x20); //JSR
    ROM.Write(0xFA1A, LevelEndMusicPatch&0xFF);
    ROM.Write(0xFA1B, LevelEndMusicPatch>>8);
    ROM.Write(0xFA1C, 0xEA); //NOP
    ROM.Write(0xFA1D+1, 0x01); // delay=1, not 0x80
#endif

#if 0 /* Make vertical scrolling faster. DOESN'T WORK YET */
    unsigned char NMICopyBuf[512];
    
    unsigned NMIoffs=0;
    #define Copy(begin,end) \
      ROM.ReadData(begin, end-begin, NMICopyBuf+NMIoffs); NMIoffs+=(end-begin);
     
    Copy(0xD4B7, 0xD4C7);
    Copy(0xD4CE, 0xD4F6);
    Copy(0xD52F, 0xD54A);
    ROM.Write(NMIoffs++, 0x60); // RTS
    #undef Copy
    
    unsigned AddrNMICopyBuf = ROM.FindFreeSpaceIn(ROMmap_npages-1, NMIoffs);
    ROM.WriteBytes(AddrNMICopyBuf, NMIoffs, NMICopyBuf);

    static const unsigned char VScrollSpeed[] =
    {
        0xA5,0x33,      //   lda $33
        0x29,0x03,      //   and #$03
        0xD0,6,         //   bne +
        0x20,0xC4,0xC6, //   jsr LoadEnemyGraphics
        
        0x20, AddrNMICopyBuf&255, AddrNMICopyBuf>>8, // jsr AddrNMICopyBuf
     //   0x20,0x1B,0xC0, //   jsr NextFrame
        0x60            // + rts 
    };
    unsigned FixVScrollSpeed = ROM.FindFreeSpaceIn(5, sizeof(VScrollSpeed));
    ROM.WriteBytes(FixVScrollSpeed, sizeof(VScrollSpeed), VScrollSpeed);
    
    ROM.Write(0x05A088, 0x20); // jsr
    ROM.Write(0x05A089, FixVScrollSpeed&0xFF); //
    ROM.Write(0x05A08A, FixVScrollSpeed >> 8); // addr
    NopRegion(0x05A08B, 3);
#endif

#if 0 /* FIX THE HORIZONTAL LEFT-SCROLLING BUG? (THIS FIX ACTUALLY WORKS) */
    static const unsigned char HScrollPatch[] =
    {
        0xAD, 0x80,0x04, //    lda MegamanPosX
        0xC9, 0x80,      //    cmp #$80
        0x90, 4,         //    bcc +
        0xA9, 0xD0,      //    lda #$E0
        0xD0, 2,         //    bne ++
        0xA9, 0x18,      // +  lda #$18
        0x8D, 0x80,0x04, // ++ sta MegamanPosX
        0x60             //    rts
    };
    unsigned FixHScrollPatch = ROM.FindFreeSpaceIn(5, sizeof(HScrollPatch));
    ROM.WriteBytes(FixHScrollPatch, sizeof(HScrollPatch), HScrollPatch);
    
    ROM.Write(0x05A054, 0x20); // jsr
    ROM.Write(0x05A055, FixHScrollPatch&0xFF); //
    ROM.Write(0x05A056, FixHScrollPatch >> 8); // addr
    NopRegion(0x05A057,2);
#endif
}
